local _, ns = ...
local OnCast = ns.addon
local Output = OnCast:NewModule("Output")

-- Lua APIs
local pcall = pcall
local pairs = pairs
local print = print
local sformat = string.format
local sgsub = string.gsub
local tinsert = table.insert
local tremove = table.remove

-- WoW APIs
local WOW = OnCast.WOWAPI or {}
WOW.SendChatMessage = SendChatMessage -- https://wow.gamepedia.com/API_SendChatMessage
WOW.IsInInstance = IsInInstance -- https://wow.gamepedia.com/API_IsInInstance
WOW.GetRaidRosterInfo = GetRaidRosterInfo -- https://wow.gamepedia.com/API_GetRaidRosterInfo
WOW.UnitInRaid = UnitInRaid -- https://wow.gamepedia.com/API_UnitInRaid
WOW.GetNumGroupMembers = GetNumGroupMembers -- https://wow.gamepedia.com/API_GetNumGroupMembers
WOW.UnitInParty = UnitInParty -- https://wow.gamepedia.com/API_UnitInParty
WOW.UnitExists = UnitExists -- https://wow.gamepedia.com/API_UnitExists
WOW.UnitIsPlayer = UnitIsPlayer -- https://wow.gamepedia.com/API_UnitIsPlayer
WOW.UnitIsFriend = UnitIsFriend -- https://wow.gamepedia.com/API_UnitIsFriend
WOW.GetUnitName = GetUnitName -- https://wow.gamepedia.com/API_GetUnitName

-- Values
local playerName = WOW.GetUnitName("player", true)

-- Constants
local TOKEN_PATTERN = "${?(%w+)}?"

local CHANNELS = {}
-- WoW Chat Channels
CHANNELS.SAY = "SAY"
CHANNELS.RAID = "RAID"
CHANNELS.PARTY = "PARTY"
CHANNELS.WHISPER = "WHISPER"
CHANNELS.EMOTE = "EMOTE"
CHANNELS.INSTANCE = "INSTANCE_CHAT"
-- Specialised channels
CHANNELS.BATTLEGROUND = "BATTLEGROUND"
CHANNELS.ARENA = "ARENA"
CHANNELS.RAID10 = "RAID10"
CHANNELS.RAID20 = "RAID20"
CHANNELS.RAIDPARTY = "RAIDPARTY"
OnCast.CHANNELS = CHANNELS

-- Channel Aliases
local STANDARD_CHANNELS = {
    ["BATTLEGROUND"] = CHANNELS.BATTLEGROUND,
    ["PVP"] = CHANNELS.BATTLEGROUND,
    ["BG"] = CHANNELS.BATTLEGROUND,
    ["ARENA"] = CHANNELS.ARENA,
    ["AR"] = CHANNELS.ARENA,
    ["RAID"] = CHANNELS.RAID,
    ["R"] = CHANNELS.RAID,
    ["RAID10"] = CHANNELS.RAID10,
    ["R10"] = CHANNELS.RAID10,
    ["RAID20"] = CHANNELS.RAID20,
    ["R20"] = CHANNELS.RAID20,
    ["RAIDPARTY"] = CHANNELS.RAIDPARTY,
    ["RPARTY"] = CHANNELS.RAIDPARTY,
    ["RP"] = CHANNELS.RAIDPARTY,
    ["PARTY"] = CHANNELS.PARTY,
    ["P"] = CHANNELS.PARTY,
    ["WHISPER"] = CHANNELS.WHISPER,
    ["W"] = CHANNELS.WHISPER,
    ["SAY"] = CHANNELS.SAY,
    ["S"] = CHANNELS.SAY,
    ["EMOTE"] = CHANNELS.EMOTE,
    ["E"] = CHANNELS.EMOTE
}
OnCast.STANDARD_CHANNELS = STANDARD_CHANNELS

-- Extended Channels
local EXTENDED_CHANNELS = {}
OnCast.EXTENDED_CHANNELS = EXTENDED_CHANNELS
local EXTENDED_CHANNEL_INDEX = {}

local function removeExtendedChannel(name)
    for i=1, #EXTENDED_CHANNEL_INDEX do
        if EXTENDED_CHANNEL_INDEX[i] == name then
            tremove(EXTENDED_CHANNEL_INDEX, i)
            EXTENDED_CHANNELS[name] = nil
            break
        end
    end
end
OnCast.RemoveExtendedChannel = removeExtendedChannel

local function addExtendedChannel(name, extension)
    if EXTENDED_CHANNELS[name] then
        removeExtendedChannel(name)
    end
    EXTENDED_CHANNELS[name] = extension
    tinsert(EXTENDED_CHANNEL_INDEX, name)
end
OnCast.AddExtendedChannel = addExtendedChannel


-- Module functions
function Output:OnInitialize()
end

function Output:OnEnable()
end

function Output:OnDisable()
end

-- Print message functions
local function printInfoMessage(msg, color)
    color = color or "ffff00"
    print(sformat("|cffffe00aOnCast|r: |cff%s%s|r", color, msg))
end
OnCast.printInfoMessage = printInfoMessage

local function printDebugMessage(msg, color)
    color = color or "ffdddddd"
    print(sformat("|cffffe00aOnCast|r: |c%s<DEBUG>%s|r", color, msg))
end
OnCast.printDebugMessage = printDebugMessage

local function printErrorMessage(msg)
    print(sformat("|cffffe00aOnCast|r: |cffff5555!ERROR! %s|r", msg))
end
OnCast.printErrorMessage = printErrorMessage

-- Check if a conditional unit matches the target
local function matchConditionalUnit(targetName, units)
    local unit = units["player"]
    if unit and targetName == unit then
        return true
    end
    unit = units["target"]
    if unit and targetName == unit then
        return true
    end
    unit = units["mouseover"]
    if unit and targetName == unit then
        return true
    end
    unit = units["focus"]
    if unit and targetName == unit then
        return true
    end
    unit = units["targettarget"]
    if unit and targetName == unit then
        return true
    end
    return false
end

-- Group number to name
local GROUP_NAMES = {[1]="Group 1",[2]="Group 2",[3]="Group 3",[4]="Group 4",[5]="Group 5",[6]="Group 6",[7]="Group 7",[8]="Group 8"}
local function getGroupToken(target)
    if not target then return nil end
    local raidNo = WOW.UnitInRaid(target)
    if raidNo then
        local _, _, sub_group = WOW.GetRaidRosterInfo(raidNo)
        return GROUP_NAMES[sub_group]
    elseif WOW.UnitInParty(target) then
        return "Party"
    end
    return target
end

local RAID_INDEX_TOKEN = {[1]="{rt1}",[2]="{rt2}",[3]="{rt3}",[4]="{rt4}",[5]="{rt5}",[6]="{rt6}",[7]="{rt7}",[8]="{rt8}"}

local _TokenTable = {}
-- Output message on up to one public channel and all assigned extended channels
function OnCast.outputSpellCast(event, cast)
    -- Select target name from cast or event
    local targetName = cast.targetName
    if event.meta.target then
        if (not targetName) or targetName == "Unknown" then
            targetName = event.meta.target.name
        elseif targetName ~= event.meta.target.name then
            return -- Discard output if we have a mis-matched target
        end
    end

    local channels = event.channels
    -- Check for conditional channels
    if event.conditionalChannels then
        channels = {} -- Start a new table so we don't pollute the event
        local channelCount = 0
        -- Copy non-conditional channels
        if event.channels then
            for channelName,channel in pairs(event.channels) do
                channels[channelName] = channel
                channelCount = channelCount + 1
            end
        end
        -- Merge conditional channels
        for i=1, #event.conditionalChannels do
            local conditional = event.conditionalChannels[i]
            local ok = true
            -- Units
            if conditional.units and (not matchConditionalUnit(targetName, conditional.units)) then
                ok = false
            end
            -- Channels OK
            if ok then
                for channelName,channel in pairs(conditional.channels) do
                    channels[channelName] = channel
                    channelCount = channelCount + 1
                end
            end
        end
        -- Exit if all channels were filtered
        if channelCount == 0 then 
            return 
        end
    end

    -- Select channel
    local publicChannel = false
    local publicMessage = nil
    
    -- Check Public Channels
    if not (event.meta and event.meta.skipStandard) then
        local inInstance, instanceType = WOW.IsInInstance()
        local isBG = (inInstance and instanceType == "pvp")
        local isArena = (inInstance and instanceType == "arena")
        -- PvP
        if isArena or isBG then
            if isArena and channels[CHANNELS.ARENA] then
                publicChannel = CHANNELS.INSTANCE
                publicMessage = channels[CHANNELS.ARENA].message
            elseif channels[CHANNELS.BATTLEGROUND] then
                publicChannel = CHANNELS.INSTANCE
                publicMessage = channels[CHANNELS.BATTLEGROUND].message
            end
        -- Raid
        elseif WOW.UnitInRaid("player") then
            local raidSize = WOW.GetNumGroupMembers()
            -- Check raid size
            if raidSize <= 10 then -- 1 to 10
                if channels[CHANNELS.RAID10] then
                    publicChannel = CHANNELS.RAID
                    publicMessage = channels[CHANNELS.RAID10].message
                end
            elseif raidSize <= 20 then -- 11 to 20
                if channels[CHANNELS.RAID20] then
                    publicChannel = CHANNELS.RAID
                    publicMessage = channels[CHANNELS.RAID20].message
                end
            end
            if (not publicChannel) then
                if channels[CHANNELS.RAID] then
                    publicChannel = CHANNELS.RAID
                    publicMessage = channels[CHANNELS.RAID].message
                elseif channels[CHANNELS.RAIDPARTY] then
                    publicChannel = CHANNELS.PARTY
                    publicMessage = channels[CHANNELS.RAIDPARTY].message
                end
            end
        -- Party
        elseif WOW.UnitInParty("player") then
            if channels[CHANNELS.PARTY] then
                publicChannel = CHANNELS.PARTY
                publicMessage = channels[CHANNELS.PARTY].message
            end
        end
        -- Other
        if not publicChannel then
            if channels[CHANNELS.WHISPER] and targetName and WOW.UnitExists(targetName) and WOW.UnitIsPlayer(targetName) and WOW.UnitIsFriend("player",targetName) then
                publicChannel = CHANNELS.WHISPER
                publicMessage = channels[CHANNELS.WHISPER].message
            elseif channels[CHANNELS.SAY] and inInstance then -- Automated Say is limited to inside instances
                publicChannel = CHANNELS.SAY
                publicMessage = channels[CHANNELS.SAY].message
            elseif channels[CHANNELS.EMOTE] then
                publicChannel = CHANNELS.EMOTE
                publicMessage = channels[CHANNELS.EMOTE].message
            end
        end
    end

    -- Exit if we don't have a public channel or extended channels
    if (not publicChannel) and (event.meta and event.meta.skipExtended) then 
        return
    end
    
    -- Table to replace message ${token}s with values
    _TokenTable["spell"] = cast.spell.name
    _TokenTable["target"] = targetName or "Unknown"
    _TokenTable["rt"] = event.meta.target and RAID_INDEX_TOKEN[event.meta.target.raidTargetIndex] or ""
    _TokenTable["event"] = cast.event
    _TokenTable["outcome"] = cast.outcome or "In Progress"
    _TokenTable["reason"] = cast.reason
    _TokenTable["group"] = getGroupToken(targetName)
    _TokenTable["delay"] = cast.spell.delay and sformat("%0.1fs", cast.spell.delay / 1000.0) or "0s"
    _TokenTable["time"] = cast.spell.castTime and sformat("%0.1fs", cast.spell.castTime / 1000.0) or "0s"
    _TokenTable["cooldown"] = cast.cooldown
    -- Extended
    _TokenTable["xspell"] = cast.xspell or ""
    _TokenTable["amount"] = cast.amount or 0
    _TokenTable["xtype"] = cast.xtype or ""
    -- Aliases
    _TokenTable["sp"] = _TokenTable["spell"]
    _TokenTable["t"] = _TokenTable["target"]
    _TokenTable["ev"] = _TokenTable["event"]
    _TokenTable["oc"] = _TokenTable["outcome"]
    _TokenTable["rsn"] = _TokenTable["reason"]
    _TokenTable["grp"] = _TokenTable["group"]
    _TokenTable["dl"] = _TokenTable["delay"]
    _TokenTable["tm"] = _TokenTable["time"]
    _TokenTable["cd"] = _TokenTable["cooldown"]
    -- Extended
    _TokenTable["xsp"] = _TokenTable["xspell"]
    _TokenTable["amt"] = _TokenTable["amount"]
    _TokenTable["xtp"] = _TokenTable["xtype"]
    
    -- Output to the public channel
    if publicChannel then
        local message = sgsub(publicMessage, TOKEN_PATTERN, _TokenTable)
        if publicChannel == CHANNELS.WHISPER then
            if targetName == playerName then -- Print info message instead of whispering yourself
                printInfoMessage(message, "ff78ea")
            else
                local ok, error = pcall(WOW.SendChatMessage, message, publicChannel, nil, targetName)
                if (not ok) then OnCast.printErrorMessage(error) end
            end
        else
            local ok, error = pcall(WOW.SendChatMessage, message, publicChannel)
            if (not ok) then OnCast.printErrorMessage(error) end
        end
    end
    
    -- Extended channels don't support raid token replacement
    _TokenTable["rt"] = ""

    -- Output to all assigned extended channels
    if not (event.meta and event.meta.skipExtended) then
        for i=1, #EXTENDED_CHANNEL_INDEX do
            local extName = EXTENDED_CHANNEL_INDEX[i]
            local channel = channels[extName]
            if channel then
                local extension = EXTENDED_CHANNELS[extName]
                local message = sgsub(channel.message, TOKEN_PATTERN, _TokenTable)
                local ok, error = pcall(extension.ShowMessage, message, channel.vars, cast)
                if (not ok) then OnCast.printErrorMessage(error) end
            end
        end
    end
end
